// MemoryMatrix.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ArrayImpl;

/**
 * The class of the memory matrixes (layered) used by the computational objects
 * included in recipes
 *
 * @author Pietro Terna
 */
public class MemoryMatrix extends SwarmObjectImpl
{

    /** the number identifying the memory matrix (starting from zero)
     *  this number represents also the posizione of this memory matrix in an
     *  array of memory matrixes created and managed by ESFrameModelSwarm */
    public int memoryMatrixNumber;
    /** the number of layers in the simulation */
    public int totalLayerNumber;
    /** layer to be used */
    public int currentLayer;
    /** is the unit sensitive to layers? */
    public boolean sensitive;
    /** the array of matrixes built one for each layer, 
     *  if layers are > 1 and sensitive==true */
    public ArrayImpl matrixArray;
    /** the memory matrix */
    public float [][] matrix;
    /** the array of shadow matrixes built one for each layer, 
     *  if layers are > 1 and sensitive==true */
    public ArrayImpl shadowMatrixArray;
    /** the shadow memory matrix */
    public float [][] shadowMatrix;
    /** the numberOfRows of matrix */
    public int numberOfRows;
    /** the numberOfCols of matrix */
    public int numberOfCols;
    /** print matrix option */
    public boolean printMatrixes;
    /** to be used in case of a unit memory matrix:
     *  number of units using this memory matrix */
    public int numberOfUnitsUsingThisMemoryMatrix;

   /**
     * the constructor for MemoryMatrix
     */
    public MemoryMatrix (Zone aZone, int n, boolean s,
			 int tln, int nor, int noc, boolean pm)
    {
	// Call the constructor for the parent class.
	super(aZone);

	init(aZone, n, s,
	     tln, nor, noc, pm);
    }

    /** managing the constructor content */
    public void init (Zone aZone, int n, boolean s,
		      int tln, int nor, int noc, boolean pm)
    {
	int i, max, ii, jj;

	// sensitivness to layers
	sensitive=s;

	// the number of layers
	totalLayerNumber=tln;

	// the number of this memory matrix
	memoryMatrixNumber=n;

	// initialize the number of units using this memory matrix
	numberOfUnitsUsingThisMemoryMatrix=0;

	// print matrixes
	printMatrixes=pm;
	// creating the matrixes, one per layer
	max=totalLayerNumber;
	if(! sensitive)max=1;

	// creating the array of matrixes (we will have one matrix 
	// for each layer), with at least one element
	matrixArray = new ArrayImpl(getZone(), max);
	shadowMatrixArray = new ArrayImpl(getZone(), max);

	numberOfRows=nor;
	numberOfCols=noc;
	for(i=0;i<max;i++){
	    matrix=new float[numberOfRows][numberOfCols];
	    matrixArray.atOffset$put(i,matrix);    

	    shadowMatrix=new float[numberOfRows][numberOfCols];
	    shadowMatrixArray.atOffset$put(i,shadowMatrix);

	    for (ii=0;ii<numberOfRows;ii++)    
		for (jj=0;jj<numberOfCols;jj++)
		    shadowMatrix[ii][jj]=0; // 'matrix' is empty
	}


    }


   /** setting a value (float) */
    public void setValue(int layer, int row, int col, float value){

	if (layer <0 || layer >= totalLayerNumber){
	    System.out.println(
		  "Invalid layer number, setting a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}
	if (row <0 || row >= numberOfRows){
	    System.out.println(
		  "Invalid row number, setting a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}
	if (col <0 || col >= numberOfCols){
	    System.out.println(
		  "Invalid column number, setting a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}

	if(  sensitive)matrix= (float [][]) matrixArray.atOffset(layer);
	if(! sensitive)matrix= (float [][]) matrixArray.atOffset(0);
	matrix[row][col] = value;
	if(  sensitive)shadowMatrix= (float [][]) 
			   shadowMatrixArray.atOffset(layer);
	if(! sensitive)shadowMatrix= (float [][]) 
			   shadowMatrixArray.atOffset(0);
	shadowMatrix[row][col] = (float) 1;
    }

   /** setting a value (double) */
    public void setValue(int layer, int row, int col, double value){

	setValue(layer, row, col, (float) value);
    }

   /** getting a value */
    public float getValue(int layer, int row, int col){

	if (layer <0 || layer >= totalLayerNumber){
	    System.out.println(
		  "Invalid layer number, getting a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}
	if (row <0 || row >= numberOfRows){
	    System.out.println(
		  "Invalid row number, getting a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}
	if (col <0 || col >= numberOfCols){
	    System.out.println(
		  "Invalid column number, getting a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}

	if(  sensitive)shadowMatrix= (float [][]) 
			   shadowMatrixArray.atOffset(layer);
	if(! sensitive)shadowMatrix= (float [][]) 
			   shadowMatrixArray.atOffset(0);
	if(shadowMatrix[row][col]==0){
	    System.out.println(
		  "Attempt of getting a non existent value in memory "+
		  "matrix # " + memoryMatrixNumber + " at row " + row +
		  " and col " + col);
	    MyExit.exit(1);
	}

	if(  sensitive)matrix= (float [][]) matrixArray.atOffset(layer);
	if(! sensitive)matrix= (float [][]) matrixArray.atOffset(0);
	return matrix[row][col];
    }

    /** checking if a position in a matrix is empty */
    public boolean getEmpty(int layer, int row, int col){

	if (layer <0 || layer >= totalLayerNumber){
	    System.out.println(
		  "Invalid layer number, checking a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}
	if (row <0 || row >= numberOfRows){
	    System.out.println(
		  "Invalid row number, checking a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}
	if (col <0 || col >= numberOfCols){
	    System.out.println(
		  "Invalid column number, "+
		  "checking a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}

	if(  sensitive)shadowMatrix= (float [][]) 
			   shadowMatrixArray.atOffset(layer);
	if(! sensitive)shadowMatrix= (float [][]) 
			   shadowMatrixArray.atOffset(0);

	if(shadowMatrix[row][col]==0) return true;  // empty!!
	else                          return false; // full!!

    }

    /** setting a position in a matrix as empty */
    public void setEmpty(int layer, int row, int col){

	if (layer <0 || layer >= totalLayerNumber){
	    System.out.println(
		  "Invalid layer number, checking a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}
	if (row <0 || row >= numberOfRows){
	    System.out.println(
		  "Invalid row number, checking a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}
	if (col <0 || col >= numberOfCols){
	    System.out.println(
		  "Invalid column number, "+
		  "checking a value in memory matrix # " 
		  + memoryMatrixNumber);
	    MyExit.exit(1);
	}

	if(  sensitive)shadowMatrix= (float [][]) 
			   shadowMatrixArray.atOffset(layer);
	if(! sensitive)shadowMatrix= (float [][]) 
			   shadowMatrixArray.atOffset(0);

	shadowMatrix[row][col]=0;  // now empty!!
    }

    /** printing the matrix on the current terminal */
    public void print(){
	int i, j, layer, max;
	if(! printMatrixes)return;


	max=totalLayerNumber;
	if(! sensitive)max=1;

	System.out.println("Memory matrix number " + memoryMatrixNumber +
			   " (rows=" + numberOfRows + 
                           ", cols=" + numberOfCols + ")");

	for (layer=0;layer<max;layer++)
	    {
		if(! sensitive)
		    System.out.println("Unlayered");
		else
		    System.out.println("Layer " + layer);

		if(  sensitive)shadowMatrix= (float [][]) 
				   shadowMatrixArray.atOffset(layer);
		if(! sensitive)shadowMatrix= (float [][]) 
				   shadowMatrixArray.atOffset(0);
		if(  sensitive)matrix= (float [][]) 
				   matrixArray.atOffset(layer);
		if(! sensitive)matrix= (float [][]) 
				   matrixArray.atOffset(0);

		for (i=0;i<numberOfRows;i++)
		    {
			for (j=0;j<numberOfCols;j++)
			    {
				if(shadowMatrix[i][j]==0)
				    System.out.print("NA ");
				else
				    System.out.print(matrix[i][j] + " ");
			    }

			System.out.println("");
		    }
	    }
	System.out.println("");
    }

    /** printing the matrix on the current terminal regardless to the
     * value of printMatrixes */
    public void printDirectly(){
	int i, j, layer, max;

	max=totalLayerNumber;
	if(! sensitive)max=1;

	System.out.println("Memory matrix number " + memoryMatrixNumber +
			   " (rows=" + numberOfRows + 
                           ", cols=" + numberOfCols + ")");

	for (layer=0;layer<max;layer++)
	    {
		if(! sensitive)
		    System.out.println("Unlayered");
		else
		    System.out.println("Layer " + layer);

		if(  sensitive)shadowMatrix= (float [][]) 
				   shadowMatrixArray.atOffset(layer);
		if(! sensitive)shadowMatrix= (float [][]) 
				   shadowMatrixArray.atOffset(0);
		if(  sensitive)matrix= (float [][]) 
				   matrixArray.atOffset(layer);
		if(! sensitive)matrix= (float [][]) 
				   matrixArray.atOffset(0);

		for (i=0;i<numberOfRows;i++)
		    {
			for (j=0;j<numberOfCols;j++)
			    {
				if(shadowMatrix[i][j]==0)
				    System.out.print("NA ");
				else
				    System.out.print(matrix[i][j] + " ");
			    }

			System.out.println("");
		    }
	    }
	System.out.println("");
    }

    /** returning the number of rows */
    public int getNumberOfRows()
    {
	return numberOfRows;
    }    

    /** returning the number of cols */
    public int getNumberOfCols()
    {
	return numberOfCols;
    }    

    /** returning the number of the matrix */
    public int getMemoryMatrixNumber()
    {
	return memoryMatrixNumber;
    }    

    /** setting the number of units using this memory matrix (++) */    
   public void increaseNumberOfUnitsUsingThisMemoryMatrix()
    {
	numberOfUnitsUsingThisMemoryMatrix++;
    }

    /** setting the number of units using this memory matrix (--) */    
   public void decreaseNumberOfUnitsUsingThisMemoryMatrix()
    {
	numberOfUnitsUsingThisMemoryMatrix--;
    }

    /** getting the number of units using this memory matrix */    
   public int getNumberOfUnitsUsingThisMemoryMatrix()
    {
	return numberOfUnitsUsingThisMemoryMatrix;
    }

    /** dropping a memory matrix */
    public void drop()
    {
	decreaseNumberOfUnitsUsingThisMemoryMatrix();

	if(getNumberOfUnitsUsingThisMemoryMatrix()<0)
	    {
		System.out.println("Internal error: the number of units using "+
				   "a unit memory matrix cannot be <0.");
		MyExit.exit(1);
	    }

	// no more units using this memory matrix
	// if(getNumberOfUnitsUsingThisMemoryMatrix()==0)super.drop();
	if(getNumberOfUnitsUsingThisMemoryMatrix()==0)
	    InformationRepository.addToHibernatedMemoryMatrixList(this);

    }

}
